module type M = sig
  type t
  type init
  type action_functional
  type action_ui
  val create : init -> t
  val on_action_func : t -> action_functional -> (t * action_ui) Lwt.t
end

module Core (M : M) = struct
  type callback = M.action_ui -> unit
  type propagator = M.t -> unit

  class core ?(propagators=[]) x =
    let e, e_send = Lwt_react.E.create () in
    let callbacks : callback Queue.t = Queue.create () in
    let f state action_func =
      lwt state', action_ui = M.on_action_func state action_func in
      Queue.iter (fun cb -> cb action_ui) callbacks;
      (if state' != state then List.iter (fun f -> f state') propagators);
      Lwt.return state'
    in
    let state_init = M.create x in
    object
      initializer
        List.iter (fun f -> f state_init) propagators
      val state = Lwt_react.E.fold_s f state_init e
      val mutable propagators : propagator list = propagators
      method propagator_add propagator =
        propagators <- propagator :: propagators
      method propagator_rm propagator =
        propagators <- List.filter ((!=) propagator) propagators
      method connect callback = Queue.push callback callbacks
      method send = e_send
    end
end
