module type Set = sig
  type 'a t
  val add : 'a t -> 'a -> 'a t
  val remove : 'a t -> 'a t
  val empty : 'a t
  val get : 'a t -> 'a
  val length : 'a t -> int
  val pos : 'a t -> int
  val nth : 'a t -> int -> 'a t
  module Iter : sig
    type 'a iter
    val to_pos : 'a t -> 'a iter -> int
    val of_pos : 'a t -> int -> 'a iter
    val of_set : 'a t -> 'a iter
    val next : 'a t -> 'a iter -> 'a iter
  end
end

module CustomListView (Set : Set) = struct
  class [ 'a ] model obj = object(self)
    inherit [ 'a Set.Iter.iter, 'a Set.Iter.iter, unit, unit ] GTree.custom_tree_model obj

    val mutable rows : 'a Set.t = Set.empty

    method custom_flags =
      [ `LIST_ONLY ]

    method insert t =
      rows <- Set.add rows t;
      let pos = Set.pos rows in
      self#custom_row_inserted (GTree.Path.create [ pos ]) (Set.Iter.of_set rows)

    method remove () =
      let pos = GTree.Path.create [ Set.pos rows ] in
      rows <- Set.remove rows;
      self#custom_row_deleted pos

    method custom_get_path iter =
      GTree.Path.create [ Set.Iter.to_pos rows iter ]

    (* data in the column number $column of row $row *)
    method custom_value g_type row ~column =
      rows <- Set.nth rows (Set.Iter.to_pos rows row);
      List.nth (Set.get rows) column

    method custom_get_iter path =
      let n = (GTree.Path.get_indices path).(0) in
      if n < 0 || n >= Set.length rows then
        None
      else
        Some (Set.Iter.of_pos rows n)

    method custom_iter_children _row =
      None

    method custom_iter_has_child _row =
      false

    method custom_iter_n_children = function
      | Some _row -> 0
      | None -> Set.length rows

    method custom_iter_next row =
      if Set.Iter.to_pos rows row >= Set.length rows - 1 then
        None
      else
        Some (Set.Iter.next rows row)

    method custom_iter_nth_child _row n =
      if 0 > n || n >= Set.length rows then
        None
      else
        Some (Set.Iter.of_pos rows n)

    method custom_iter_parent _row =
      None

    method custom_encode_iter b =
      b, (), ()

    method custom_decode_iter b () () =
      b
  end
end

